unit SIP;

// SIP Client for Delphi 6PE, (c)2006 by Paul TOTH <tothpaul@free.fr>
// http://tothpaul.free.fr

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

(* Documentation :
 ---------------

 Session Initiation Protocol (RFC 3261) :

  http://fr.wikipedia.org/wiki/Session_Initiation_Protocol
  ftp://ftp.rfc-editor.org/in-notes/rfc3261.txt

 Session Description Protocol (RFC 2327) :

 Other links :

  http://abcdrfc.free.fr/ [RFC en franais]


 Historique :
 ------------

 27 septembre 2006 - 0.4.0 (Paul TOTH)
  - validation du code de Fritivi
  - modification de l'interface graphique

 20 septembre 2006 - x.x.x (by Fritivi - http://fritivi.free.fr)
  - rception des appels entrants
  - modification de l'historique des appels : ajout du type d'appel, sparation des colonnes avec tabulation

 11 septembre 2006 - 0.3.3 (Paul TOTH)
  - correction d'un erreur dans l'ordre d'appel de Start et Step:=csONLINE
  - ajout de l'historique des 15 derniers appels

 30 aot 2006 - 0.3.2  (Paul TOTH)
  - merci  Nicolas qui m'a ouvert les yeux !
  - la session audio doit dmarrer ds que possible (document SDP), ce qui peut se
    produire lors du TRYING, donc bien que le correspondant dcroche.
    Du coup, on a maintenant la sonnerie :D

 28 aot 2006 - 0.3.1 (Paul TOTH)
  - ajout du check de version (dans main.pas)
  - utilise le MD5 par dfault (rend SIPInside compatible avec sip.1und1.de).

 11 aot 2006 - 0.3.0 (Paul TOTH)
  - passage en mode multithread
  - utilisation de REGISTER

 11 aot 2006 - 0.2.3 (Paul TOTH)
  - petit bug de Crunch()

 5 aot 2006 - 0.2.2 (Paul TOTH)
  - codage du mot de passe, le calcul A1 peut remplacer le mot de passe sous la forme {A1}

 1er aot 2006 (Paul TOTH)
 - premire release des sources

 17 juillet 2006 - b0.1.2 :
  - rcriture partie UDP

 12 juillet 2006 - b0.1.1 :
  - passage en mode asynchrone
  - avec gestion de l'tat de la "communication" avec OnStep
  - gestion de l'interactivit (SIP.Cancel)

  NB: il faut revoir tout l'aspect socket I/O car en UDP a ne devrait pas fonctionner comme je l'ai fait

 11 juillet 2006 - b0.1 :
  - premiers tests en mode bloquant, a fonctionne en UDP, mais il reste la partie RTP/AVP !

 SDP (RFC 2327)

*)

interface

uses
 Windows, Messages, Winsock,
 Classes, SysUtils, UDP;

const
 SIP_VERSION  = 'SIPInside 0.4';
 SIP_RELEASE  = SIP_VERSION+' (2006-09-27)';

 SIP_DEFAULT_PORT = 5060;

// SIP Error
 SIP_NO_ERROR        = 0;
 SIP_HOST_NOT_FOUND  = 1;
 SIP_NO_SOCKET       = 2;
 SIP_SEND_ERROR      = 3;
 SIP_RECV_ERROR      = 4;
 SIP_INVALID_RESPONSE= 5;
 SIP_UNAUTHORIZED    = 6;
 SIP_NOT_CONNECTED   = 7;

 SIP_RECVDATA        = 1;

type
 TSIPStep=(
  csNONE,       // before or after a conversation
  csREGISTER,   // send REGISTER
  csREGISTERED, // Registered
  csINVITE,     // send INVITE
  csAUTH,       // 40x Authenticate
  csTRYING,     // 101 Trying
  csWRONG,      // 403 Wrong login or password
  csRINGING,    // 108 Ringing
  csCANCEL,     // send CANCEL
  csCANCELED,   // 487 Canceled, or cancel INVITE (no response from server)
  csUNAVAILABLE,// 480 Temporarily Unavailable
  csONLINE,     // 200 OK after INVITE, TRYING or RINGING
  csSENDBYE,    // send BYE
  csRECVBYE,     // BYE received

  csRECVINVITE,  // Incoming call !
  csCALLREFUSED, // Don't talk to me ! => 486 busy here
  csCALLACCEPTED,// OK envoy,
  csRECVCANCEL   // Caller has cancelled the call...
 );

 TSIPMode=(
  smLOGIN,
  smCALL,
  smLOGOUT
 );

 TSIPCallType=(
  ctIN,
  ctOUT,
  ctMISSED
 );

 TCustomSIPClient=class(TCustomUDPClient)
 private
  fHWnd      :THandle;   // Window Handle used for Async sockets
  fMsg       :integer;   // Message for Async sockets
  fActive    :boolean;
  fThread    :THandle;
 private
  fSession   :integer;   // SDP session
 // Socket
  fRemote    :TSockAddr; // SockAddr to send data to fDomain
  fReceived  :string;    // fDomain IP address
  fExpires   :string;    // 1800 or 0 for Logout
  fVia       :string;    // SIP Via header line
  fContact   :string;    // SIP Contact header line
  fBranch    :string;    // SIP Via header line Branch value
  fFrom      :string;    // SIP From header line
  fCallID    :string;    // SIP Call-ID
  fRequest   :string;    // last Request
  fURI       :string;    // URI
  fTarget    :string;    // Target
  fTagged    :string;    // Tagged target
  fSequence  :cardinal;  // Request sequence
  fStep      :TSIPStep; // Phone call step
  fMode      :TSIPMode;
  fCallType  :TSIPCallType;
 // socket I/O
  procedure WriteData;
 // extract header info
  function IntegerAt(Offset:integer):integer;
  function getHeader(const Tag:string):string;
 // generic send request method
  procedure SendRequest(const Request,Target,Headers,Content:string);
 // send a request with a new Request number
  procedure ExecRequest(const Request,Target,Headers,Content:string);
 // retreive SIP status code
  function GetStatus:integer;
 // send a status code for a SIP request (BYE)
  procedure SendReply(const Status:string);
  procedure SendSDPReply(const Status:string);
 // compute authorisation response (RFC 2617)
  function GetAuthorization(const Method,URI:string):string;
 // resend request with authorization response
  procedure SendAuthorization;
  procedure ThreadProc;
 protected
  fName      :string;    // User name
  fUserID    :string;    // SIP userid
  fPasswd    :string;    // SIP password
  fPhone     :string;    // Phone number to call -> <sip:fPhone@fDomain>
  fDomain    :string;    // SIP domain
  fData      :string;    // request body
  fSend      :string;    // send buffer
  fHead      :string;    // request header
  fStatus    :integer;   // request status code
  fLocalSDP  :string;
  fLocalTag  :string;
  fRemoteSDP :string;
  fRegistered:boolean;
  fAccept    :boolean;
  procedure RecvStatus;
  procedure RecvRequest;
  procedure RecvData; virtual;
  procedure SendData; virtual;
  procedure Open; virtual;   // REGISTERED
  procedure Start; virtual;  // dbut de conversation
  procedure Stop; virtual;   // fin de communication
  procedure SetStep(Value:TSIPStep); virtual;
  property Error:integer read fError write SetError;
  property Step:TSIPStep read fStep write SetStep;
  property CallType:TSIPCallType read fCallType write fCallType;
 public
  procedure Close; override; // UNREGISTER, close socket
  constructor Create(AHandle:HWnd; AMsg:integer);
  procedure SetLocal(const Addr:string; Port:word);
  procedure Login(const AUsername, ADomain, AUserID, APassword:string);
  procedure Logout;
  procedure Dial(const APhoneNumber:string);
  procedure AnswerCall;
  procedure RefuseCall;
  function Crunch(const UserID, Realm, Password:string):string;
  procedure Cancel;
  procedure ProcessMessage(var Msg);
  property Accept:boolean read fAccept write fAccept;
 end;

 TSIPClient=class(TCustomSIPClient)
 private
  EOnSend :TNotifyEvent;
  EOnRecv :TNotifyEvent;
  EOnStep :TNotifyEvent;
  EOnOpen :TNotifyEvent;
  EOnClose:TNotifyEvent;
  EOnError:TNotifyEvent;
  EOnStart:TNotifyEvent;
  EOnStop :TNotifyEvent;
 protected
  procedure RecvData; override;
  procedure SendData; override;
  procedure Open; override;
  procedure Start; override;
  procedure Stop; override;
  procedure SetStep(Value:TSIPStep); override;
 public
  procedure Close; override;
  property Name:string read fName;
  property UserID:string read fUserID;
  property Domain:string read fDomain;
  property Phone:string read fPhone;
  property Registered:boolean read fRegistered;
  property Error;
  property Step;
  property CallType;
  property Status:integer read fStatus;
  property Head:string read fHead;
  property Data:string read fData;
  property Send:string read fSend;
  property LocalSDP:string read fLocalSDP write fLocalSDP;
  property RemoteSDP:string read fRemoteSDP;
  property OnSend:TNotifyEvent read EOnSend write EOnSend;
  property OnRecv:TNotifyEvent read EOnRecv write EOnRecv;
  property OnStep:TNotifyEvent read EOnStep write EOnStep;
  property OnOpen:TNotifyEvent read EOnOpen write EOnOpen;
  property OnClose:TNotifyEvent read EOnClose write EOnClose;
  property OnStart:TNotifyEvent read EOnStart write EOnStart;
  property OnStop:TNotifyEvent read EOnStop write EOnStop;
  property OnError:TNotifyEvent read EOnError write EOnError;
 end;

implementation

uses md5;

function getValue(const key,str:string):string;
var
 i,j:integer;
begin
 Result:='';
 i:=pos(key,str);
 if i=0 then exit;
 inc(i,length(key));
 if i+2>length(str) then exit;
 if str[i]<>'=' then exit;
 inc(i);
 if str[i]='"' then begin
  inc(i);
  j:=i;
  while (j<=length(str)) and (str[j]<>'"') do inc(j);
 end else begin
  j:=i;
  while (j<=length(str))and(not (str[j] in [' ',','])) do inc(j);
 end;
 Result:=copy(str,i,j-i);
end;

function StartMe(Sender:TCustomSIPClient):integer; stdcall;
begin
 try
  Sender.ThreadProc;
  Result:=0;
 except
  Result:=-1;
 end;
end;

//----------------------------------------------------------------------------//

constructor TCustomSIPClient.Create(AHandle:HWnd; AMsg:integer);
begin
 inherited Create;
 fHwnd  :=AHandle;
 fMsg   :=AMsg;
 fCallID:=CreateID;
 fSession:=getTickCount;
end;

procedure TCustomSIPClient.SetLocal(const Addr:string; Port:word);
begin
 fLocalSDP:='v=0'#13#10+
            'o=SIPInside '+IntToStr(fSession)+' '+IntToStr(getTickCount)+' IN IP4 '+Addr+#13#10+
            's=SIP Call'#13#10+
            'c=IN IP4 '+Addr+#13#10+
            't=0 0'#13#10+
            'm=audio '+IntToStr(Port)+' RTP/AVP 8'#13#10+
            'b=AS:64'#13#10+
            'a=rtpmap:8 PCMA/8000/1'#13#10+
            'a=ptime:30'#13#10;
end;

procedure TCustomSIPClient.Login(const AUsername, ADomain, AUserID, APassword:string);
var
 ad:integer;
 id:cardinal;
begin
 Close;
// Parms
 fName  :=AUsername;
 fDomain:=ADomain;
 fUserID:=AUserID;
 fPasswd:=APassword;
// Socket
 fHead:='';
 fData:='';
// find host address
 ad:=INetAddr(fDomain);
 if ad=INADDR_NONE then begin
  Cancel;
  FatalError(SIP_HOST_NOT_FOUND);
 end;
 fReceived:=inet_ntoa(in_addr(ad));
 // open a socket
 OpenSocket;
 BindSocket(SIP_DEFAULT_PORT);
 IsMultiThread:=True;
 fActive:=True;
 fThread:=CreateThread(nil,0,@StartMe,pointer(Self),0,id);
 // call server
 FillChar(fRemote,SizeOf(fRemote),0);
 fRemote.sin_family:=AF_INET;
 fRemote.sin_port:=htons(SIP_DEFAULT_PORT);
 fRemote.sin_addr.S_addr:=ad;
// Setup request header
 fExpires:='1800';
 fVia    :='Via: SIP/2.0/UDP '+fDomain+';rport;branch=z9hG4bK';
 fFrom   :='From: '+fName+' <sip:'+fUserID+'@'+fDomain+'>;tag='+IntToStr(GetTickCount)+#13#10;
 fContact:='Contact: '+fName+' <sip:'+fUserID+'@'+fDomain+'>'#13#10;
 fRegistered:=False;
 fTagged:='';
 fStep:=csNONE;
 fMode:=smLOGIN;
 Step:=csREGISTER;
 fURI:='sip:'+fDomain;
 fTarget:=fName+' <sip:'+fUserID+'@'+fDomain+'>';
 ExecRequest('REGISTER',fName+' <sip:'+fUserID+'@'+fDomain+'>','','');
end;

procedure TCustomSIPClient.Logout;
begin
 if fRegistered=False then exit;
 fExpires:='0';
 Step:=csREGISTER;
 fURI:='sip:'+fDomain;
 fTarget:=fName+' <sip:'+fUserID+'@'+fDomain+'>';
 fMode:=smLOGOUT;
 ExecRequest('REGISTER',fName+' <sip:'+fUserID+'@'+fDomain+'>','','');
end;

procedure TCustomSIPClient.Dial(const APhoneNumber:string);
begin
 if not (Step in [csREGISTERED,csCANCELED,csRECVBYE]) then exit;
 fPhone :=APhoneNumber;
 if pos('@',fPhone)=0 then
  fURI   :='sip:'+fPhone+'@'+fDomain
 else begin
  if (copy(fPhone,1,4)<>'sip:') then fURI:='sip:'+fPhone else fURI:=fPhone;
 end;
// fTarget:=AName+' <'+fURI+'>'; todo: AName charset encoding
 fTarget:='<'+fURI+'>';
 fTagged:='';
 Step:=csINVITE;
 fMode:=smCALL;
 fRemoteSDP:='';
 fCallType:=ctOUT;
 ExecRequest('INVITE',fTarget,'Content-Type: application/sdp'#13#10,fLocalSDP);
end;

procedure TCustomSIPClient.AnswerCall;
begin
 if not (Step in [csRECVINVITE]) then exit;
 Step:=csCALLACCEPTED;
 fLocalTag:=IntToStr(getTickCount);
 SendSDPReply('200 OK');
 Step:=csONLINE;
end;

procedure TCustomSIPClient.RefuseCall;
begin
 if not (Step in [csRECVINVITE]) then exit;
 fLocalTag:=IntToStr(getTickCount);
 Step:=csCALLREFUSED;
 SendSDPReply('486 Busy here');
 Cancel;
end;


procedure TCustomSIPClient.Cancel;
begin
 case fStep of
  csUNAVAILABLE : begin
   Step:=csREGISTERED;
  end;
  csNONE, csREGISTER, csINVITE : begin
   Step:=csCANCELED;
   Stop;
  end;
  csAUTH, csTRYING, csRINGING : begin
   Step:=csCANCEL;
  // use SendRequest to keep the same Request number
   SendRequest('CANCEL',fTarget,'','');
  end;
  csONLINE : begin
   Step:=csSENDBYE;
  // use ExecRequest to provide a new Request number
   ExecRequest('BYE',fTagged,'','');
  end;
  csREGISTERED : Logout;
  csCALLREFUSED,csRECVCANCEL : begin
   Step:=csREGISTERED;
  end;
  else Stop;
 end;
end;

procedure TCustomSIPClient.Close;
begin
 if fRegistered then Logout;
 fActive:=False;
 WaitOrKill(fThread,10);
 inherited;
end;

procedure TCustomSIPClient.ThreadProc;
var
 head,i,len:integer;
begin
 try
  while fActive do begin
   fData:=Read;
   // Cirpack KeepAlive Packet ?!
   if copy(fData,1,24)='Cirpack KeepAlive Packet' then continue;
   // get Header first
   head:=pos(#13#10#13#10,fData);
   if head=0 then continue; // no header
   fHead:=copy(fData,1,head+1);
   i:=pos('Content-Length: ',fHead);
   if i=0 then continue;//FatalError(SIP_INVALID_RESPONSE);
   len:=IntegerAt(i+16);
   // now content
   if Length(fData)<head+len then continue; // need more data
   fData:=copy(fData,head+4,len);
   SendMessage(fHWnd,fMsg,SIP_RECVDATA,0);
  end;
 finally
  CloseHandle(fThread);
  fThread:=0;
 end;
end;

procedure TCustomSIPClient.ProcessMessage(var Msg);
begin
 with TMessage(Msg) do begin
  case wParam of
   SIP_RECVDATA : RecvData;
  end;
 end;
end;

procedure TCustomSIPClient.WriteData;
var
 e,len:integer;
begin
 len:=Length(fSend);
 while len>0 do begin
  e:=sendto(fSocket,fSend[1],len,0,fRemote,SizeOf(fRemote));
  if e=0 then FatalError(SIP_SEND_ERROR);
  if e<0 then begin
   e:=WSAGetLastError;
   if e=WSAEWOULDBLOCK then break;
   FatalError(e);
  end;
  if e=len then begin
   len:=0;
   fSend:='';
  end;
 end;
end;

procedure TCustomSIPClient.RecvData;
begin
 if copy(fHead,1,8)='SIP/2.0 ' then RecvStatus else RecvRequest;
end;

procedure TCustomSIPClient.RecvRequest;
var
 i:integer;
begin
 // BYE from the remote client ?
 if copy(fHead,1,4)='BYE ' then begin
  SendReply('200 OK');
  if fStep=csONLINE then begin
   Step:=csRECVBYE;
   Stop;
  end;
 end else
 if copy(fHead,1,7)='CANCEL ' then begin
  SendReply('200 OK');
  if fStep=csRECVINVITE then begin
   fCallType:=ctMISSED;
   Step:=csRECVCANCEL;
   Stop;
  end;
 end else
 if copy(fHead,1,7)='INVITE ' then begin
  if (fAccept)and(Step in [csREGISTERED,csCANCELED,csRECVBYE,csRECVCANCEL]) then begin
   SendReply('180 RINGING');
   Start;
   fTarget:=getHeader('From');
   fPhone:=getHeader('From');
   i:=pos(';',fPhone);
   if i>0 then SetLength(fPhone,i-1);
   if fPhone[length(fPhone)]<>'>' then fPhone:=fPhone+'>';
   fCallType:=ctIN;
   Step:=csRECVINVITE;
end;
 end;
end;

procedure TCustomSIPClient.RecvStatus;
begin
 fTagged:=getHeader('To');
 if fTagged='' then fTagged:=fTarget;
 case GetStatus of
 // 100 Trying
  100: if fStep in [csREGISTER, csINVITE, csAUTH] then begin
        Step:=csTRYING;
        if fData<>'' then Start; // received remote SDP, start listen
       end;
 // 180 Ringing
  180: if fStep in [csINVITE, csAUTH, csTRYING] then begin
        Step:=csRINGING;
        if (fRemoteSDP='')and(fData<>'') then Start;
       end;
 // 200 OK
  200: case fStep of
        csREGISTER,
        csINVITE,
        csAUTH,
        csTRYING,
        csRINGING : begin
         case fMode of
          smLOGIN: begin
           Step:=csREGISTERED;
           Open;
          end;
          smCALL: begin
           if fRemoteSDP='' then Start; // place cette ligne avec le changement de Step
           Step:=csONLINE;              // pour un pb de squence PCMA.Listen() et PCMA.OpenIn;
           SendRequest('ACK',fTagged,'','');
          end;
          smLOGOUT: begin
           fRegistered:=False;
           Step:=csNONE;
           Close;
          end;
         end;
        end;
        csSENDBYE: begin
         Step:=csREGISTERED;
         Stop;
        end;
       end;
 // 403 Wrong login or password
  403: begin
        Step:=csWRONG;
        SendRequest('ACK',fTagged,'',''); // WRONG
        Stop;
       end;
 // 401 WWW-Authenticate
 // 407 Proxy-Authenticate
  401,
  407: if fStep in [csREGISTER,csINVITE] then begin
        Step:=csAUTH;
        SendAuthorization;
       end;
 // 480 Temporarily Unavailable
  480: begin
        Step:=csUNAVAILABLE;
        SendRequest('ACK',fTagged,'',''); // WRONG
       end;
 // 487 Canceled
  487: if fStep=csCANCEL then begin
        Step:=csCANCELED;
        SendRequest('ACK',fTagged,'',''); // CANCELED
        Stop;//Close;
       end;
 end;
end;

procedure TCustomSIPClient.Open;
begin
 fRegistered:=True;
end;

procedure TCustomSIPClient.Start;
begin
// fData = remote SDP
 if (Step<>csCALLACCEPTED) then fRemoteSDP:=fData;
end;

procedure TCustomSIPClient.Stop;
begin
// if fRegistered then Step:=csREGISTERED else Step:=csNONE;
end;

procedure TCustomSIPClient.SendData;
begin
 if fSocket=INVALID_SOCKET then FatalError(SIP_NOT_CONNECTED);
 WriteData;
end;

function TCustomSIPClient.IntegerAt(Offset:integer):integer;
var
 num:char;
begin
 Result:=0;
 while (Offset<=Length(fHead)) do begin
  num:=fHead[Offset];
  if num in ['0'..'9'] then
   Result:=10*Result+ord(num)-ord('0')
  else
   exit;
  inc(Offset);
 end;
end;

function TCustomSIPClient.getHeader(const Tag:string):string;
var
 i,j:integer;
begin
 Result:='';
 i:=pos(#13#10+Tag+':',fHead);
 if i=0 then exit;
 inc(i,2+Length(Tag)+1);
 j:=i;
 while (j<=Length(fHead))and(fHead[j]<>#13) do inc(j);
 Result:=copy(fHead,i,j-i);
end;

procedure TCustomSIPClient.SendRequest(const Request,Target,Headers,Content:string);
begin
 fRequest:=Request;
 fSend:=Request+' '+fURI+' SIP/2.0'#13#10+
        fVia+fBranch+#13#10+
        fFrom+
        'To: '+Target+#13#10+
        'Call-ID: '+fCallID+'@'+fDomain+#13#10+
        'CSeq: '+IntToStr(fSequence)+' '+Request+#13#10+
        'Max-Forwards: 70'#13#10+
        Headers+
        'Content-Length: '+IntToStr(Length(Content))+#13#10+
        #13#10+
        Content;
 SendData;
end;

function TCustomSIPClient.GetStatus:integer;
begin
 if (copy(fHead,1,8)<>'SIP/2.0 ') then FatalError(SIP_INVALID_RESPONSE);
 fStatus:=IntegerAt(9);
 if (fStatus<100)or(fStatus>699)  then FatalError(SIP_INVALID_RESPONSE);
 Result:=fStatus;
end;

procedure TCustomSIPClient.SendReply(const Status:string);
begin
 fSend:='SIP/2.0 '+Status+#13#10+
        'Via:'+getHeader('Via')+';received='+fReceived+#13#10+
        'From:'+getHeader('From')+#13#10+
        'To:'+getHeader('To')+#13#10+
        'Call-ID:'+getHeader('Call-ID')+#13#10+
        'CSeq:'+getHeader('CSeq')+#13#10+
        'Content-Length: 0'#13#10+
        #13#10;
 SendData;
end;

procedure TCustomSIPClient.SendSDPReply(const Status:string);
begin
 fSend:='SIP/2.0 '+Status+#13#10+
        'Via:'+getHeader('Via')+';received='+fReceived+#13#10+
        'From:'+getHeader('From')+#13#10+
        'To:'+getHeader('To')+';tag='+fLocalTag+#13#10+
        'Call-ID:'+getHeader('Call-ID')+#13#10+
        'CSeq:'+getHeader('CSeq')+#13#10+
        fContact+#13#10+
        'Content-Type: application/sdp'#13#10+
        'Content-Length: '+IntToStr(Length(fLocalSDP))+#13#10+
        #13#10+
        fLocalSDP;
 SendData;
end;

procedure TCustomSIPClient.ExecRequest(const Request,Target,Headers,Content:string);
begin
 fBranch:=CreateID;
 inc(fSequence);
 SendRequest(
  Request,Target,
  fContact+
 'Expires: '+fExpires+#13#10+
 'User-Agent: '+SIP_RELEASE+#13#10+
  Headers,
  Content
 );
end;

procedure TCustomSIPClient.SendAuthorization;
var
 reqs:string;
 auth:string;
begin
 reqs:=fRequest;
 auth:=GetAuthorization(fRequest,fURI);
 if reqs='INVITE' then begin
  SendRequest('ACK',fTagged,'','');
  ExecRequest(reqs,fTarget,auth+'Content-Type: application/sdp'#13#10,fLocalSDP)
 end else begin
  ExecRequest(reqs,fTarget,auth,'');
 end;
end;

function TCustomSIPClient.Crunch(const UserID, Realm, Password:string):string;
begin
 if (Length(Password)=2+32)and(Password[1]='{')and(Password[34]='}') then
  Result:=copy(Password,2,32)
 else
  Result:=MD5Print(MD5String(UserID+':'+Realm+':'+Password));
end;

function TCustomSIPClient.GetAuthorization(const Method,URI:string):string; // RFC 2617
var
 auth  :string;
 algo  :string;
 realm :string;
 nonce :string;
 opaque:string;
 A1,A2 :string;
 resp  :string;
begin
 Result:='';
 if fStatus=407 then begin
  auth:=getHeader('Proxy-Authenticate');
  if auth='' then FatalError(SIP_UNAUTHORIZED);
  Result:='Proxy-';
 end else begin
  auth:=getHeader('WWW-Authenticate');
  if auth='' then FatalError(SIP_UNAUTHORIZED);
 end;

 algo:=getValue('algorithm',auth);
// MD5 is the default algorithm, and now SIPInside work with most popular
// SIP provider in Germany sip.1und1.de :D
 if (algo<>'')and(algo<>'MD5') then FatalError(SIP_UNAUTHORIZED);

 realm :=getValue('realm' ,auth);
 nonce :=getValue('nonce' ,auth);
 opaque:=getValue('opaque',auth);

 A1:=Crunch(fUserID,realm,fPasswd);
 A2:=MD5Print(MD5String(Method+':'+URI));
 resp:=MD5Print(MD5String(A1+':'+nonce+':'+A2));

 Result:=Result+'Authorization: Digest username="'+fUserID+'",realm="'+realm+'",nonce="'+nonce+'",response="'+resp+'",uri="'+URI+'",algorithm=MD5,opaque="'+opaque+'"'#13#10;
end;

procedure TCustomSIPClient.SetStep(Value:TSIPStep);
begin
 fStep:=Value;
end;

//----------------------------------------------------------------------------//
procedure TSIPClient.RecvData;
begin
 if Assigned(EOnRecv) then EOnRecv(Self);
 inherited;
end;

procedure TSIPClient.SendData;
begin
 if Assigned(EOnSend) then EOnSend(Self);
 inherited;
end;

procedure TSIPClient.Open;
begin
 inherited;
 if Assigned(EOnOpen) then EOnOpen(Self);
end;

procedure TSIPClient.Start;
begin
 inherited;
 if Assigned(EOnStart) then EOnStart(Self);
end;

procedure TSIPClient.Stop;
begin
 inherited;
 if Assigned(EOnStop) then EOnStop(Self);
end;

procedure TSIPClient.SetStep(Value:TSIPStep);
begin
 inherited;
 if Assigned(EOnStep) then EOnStep(Self)
end;

procedure TSIPClient.Close;
begin
 if (fSocket<>INVALID_SOCKET) and Assigned(EOnClose) then EOnClose(Self);
 inherited;
end;

//----------------------------------------------------------------------------//

end.

